﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections.Generic;
#if USE_FARSEER
using FarseerGames.FarseerPhysics.Mathematics;
#endif
using System.Text;

namespace SilverArcade.SilverSprite.Graphics
{
    public class SpriteBatchCanvas
    {
        public Canvas Canvas;
        bool visible = true;
        Dictionary<long, SpriteList> clippedImages;
        Dictionary<long, SpriteList> texts;
        int zIndex;

        Matrix transMatrix = Matrix.Identity;
        System.Windows.Media.MatrixTransform transform = new System.Windows.Media.MatrixTransform();

        public bool InUse
        {
            get;
            set;
        }

        public SpriteBatchCanvas()
        {
            Canvas = new Canvas();
            transform.Matrix = System.Windows.Media.Matrix.Identity;
            Canvas.RenderTransform = transform;
            clippedImages = new Dictionary<long, SpriteList>();
            texts = new Dictionary<long, SpriteList>();
        }

        public void End()
        {
            foreach (SpriteList dl in clippedImages.Values)
            {
                dl.EndSpriteBatch();
            }
            foreach (SpriteList dl in texts.Values)
            {
                dl.EndSpriteBatch();
            }
        }

        public void Begin()
        {
            foreach (SpriteList dl in clippedImages.Values)
            {
                dl.BeginSpriteBatch();
            }
            foreach (SpriteList dl in texts.Values)
            {
                dl.BeginSpriteBatch();
            }
            zIndex = 0;
        }

        public bool Visible
        {
            set
            {
                if (value != visible)
                {
                    visible = value;
                    if (visible == false)
                    {
                        Canvas.Visibility = Visibility.Collapsed;
                    }
                    else
                    {
                        Canvas.Visibility = Visibility.Visible;
                    }
                }
            }
            get
            {
                return visible;
            }
        }

        bool MatricesAreEqual(Matrix m1, Matrix m2)
        {
            if (m1.M11 != m2.M11) return false;
            if (m1.M12 != m2.M12) return false;
            if (m1.M13 != m2.M13) return false;
            if (m1.M14 != m2.M14) return false;
            if (m1.M21 != m2.M21) return false;
            if (m1.M22 != m2.M22) return false;
            if (m1.M23 != m2.M23) return false;
            if (m1.M24 != m2.M24) return false;
            if (m1.M31 != m2.M31) return false;
            if (m1.M32 != m2.M32) return false;
            if (m1.M33 != m2.M33) return false;
            if (m1.M34 != m2.M34) return false;
            if (m1.M41 != m2.M41) return false;
            if (m1.M42 != m2.M42) return false;
            if (m1.M43 != m2.M43) return false;
            if (m1.M44 != m2.M44) return false;
            return true;
        }

        public Matrix TransformMatrix
        {
            get
            {
                return transMatrix;
            }
            set
            {
                if (MatricesAreEqual(transMatrix, value) == false)
                {
                    System.Windows.Media.Matrix t = new System.Windows.Media.Matrix(value.M11, value.M12, value.M21, value.M22, value.M41, value.M42);
                    transform.Matrix = t;
                    transMatrix = value;
                }
            }
        }

        internal void Draw(DrawCommand cmd)
        {
                switch (cmd.CommandType)
                {
                    case DrawCommand.DrawCommandType.String:
                        DrawString(cmd.Font, cmd.Text, cmd.Position, cmd.Color, cmd.Rotation, cmd.Origin, cmd.Scale, cmd.Effects, cmd.LayerDepth);
                        break;
                    case DrawCommand.DrawCommandType.Texture:
                        if (float.IsInfinity(cmd.Scale.X)) cmd.Scale.X = 0;
                        if (float.IsInfinity(cmd.Scale.Y)) cmd.Scale.Y = 0;
                        Draw(cmd.Texture, cmd.Position, cmd.SourceRectangle, cmd.Color, cmd.Rotation, cmd.Origin, cmd.Scale, cmd.Effects, cmd.LayerDepth);
                        break;
                }
        }

        public void DrawString(SpriteFont spriteFont, StringBuilder text, Vector2 position, Color color)
        {
            DrawString(spriteFont, text.ToString(), position, color);
        }

        public void DrawString(SpriteFont spriteFont, string text, Vector2 position, Color color, float rotation, Vector2 origin, Vector2 scale, SpriteEffects effects, float layerDepth)
        {
            if (text == null) return;
            if (spriteFont is BitmapSpriteFont)
            {
                ((BitmapSpriteFont)spriteFont).Draw(this, text, position, color, rotation, origin, scale, effects, layerDepth);
                return;
            }
            SpriteList l;
            long key = spriteFont.AssetId;
            if (texts.ContainsKey(key))
            {
                l = texts[key];
            }
            else
            {
                l = new SpriteList(this, SpriteList.SpriteType.SpriteText);
                l.SpriteFont = spriteFont;
                l.ParentCanvas = Canvas;
                texts.Add(key, l);
            }
            SpriteText t = l.GetSprite() as SpriteText;
            if (t.ZIndex < zIndex)
            {
                t.ZIndex = zIndex;
            }
            else
            {
                zIndex = t.ZIndex;
            }
            zIndex++;
            t.InUse = true;

            t.Position = position;
            t.Color = color;
            t.Text = text;
            t.Scale = scale;
            t.Rotation = rotation;
            t.Origin = origin;
            t.Effects = effects;
        }

        public void DrawString(SpriteFont spriteFont, StringBuilder text, Vector2 position, Color color, float rotation, Vector2 origin, Vector2 scale, SpriteEffects effects, float layerDepth)
        {
            DrawString(spriteFont, text.ToString(), position, color, rotation, origin, scale, effects, layerDepth);
        }

        public void DrawString(SpriteFont spriteFont, StringBuilder text, Vector2 position, Color color, float rotation, Vector2 origin, float scale, SpriteEffects effects, float layerDepth)
        {
            DrawString(spriteFont, text.ToString(), position, color, rotation, origin, scale, effects, layerDepth);
        }

        public void DrawString(SpriteFont spriteFont, string text, Vector2 position, Color color, float rotation, Vector2 origin, float scale, SpriteEffects effects, float layerDepth)
        {
            if (text == null) return;
            if (spriteFont is BitmapSpriteFont)
            {
                ((BitmapSpriteFont)spriteFont).Draw(this, text, position, color, rotation, origin, scale, effects, layerDepth);
                return;
            }
            SpriteList l;
            long key = spriteFont.AssetId;
            if (texts.ContainsKey(key))
            {
                l = texts[key];
            }
            else
            {
                l = new SpriteList(this, SpriteList.SpriteType.SpriteText);
                l.SpriteFont = spriteFont;
                l.ParentCanvas = Canvas;
                texts.Add(key, l);
            }
            SpriteText t = l.GetSprite() as SpriteText;
            if (t.ZIndex < zIndex)
            {
                t.ZIndex = zIndex;
            }
            else
            {
                zIndex = t.ZIndex;
            }
            Vector2 vscale = new Vector2(scale, scale);

            zIndex++;
            t.InUse = true;
            t.Position = position;
            t.Color = color;
            t.Text = text;
            t.Scale = vscale;
            t.Rotation = rotation;
            t.Origin = origin;
            t.Effects = effects;
        }

        public void DrawString(SpriteFont font, string text, Vector2 position, Color color)
        {
            DrawString(font, text, position, color, 0, Vector2.Zero, 1, SpriteEffects.None, 0);
        }


        ClippedSpriteImage GetClippedSprite(Texture2D texture, Color color)
        {
            SpriteList l;
            long key = texture.AssetId + (long)((((color.R * 256) + color.G * 256) + color.B)) << 32;


            if (clippedImages.ContainsKey(key))
            {
                l = clippedImages[key];
            }
            else
            {
                l = new SpriteList(this, SpriteList.SpriteType.ClippedSpriteImage);
                if (color.R != 255 || color.G != 255 || color.B != 255)
                {
                    l.Texture2D = new Texture2D(texture, color);
                }
                else
                {
                    l.Texture2D = texture;
                }
                l.ParentCanvas = Canvas;
                clippedImages.Add(key, l);
            }
            ClippedSpriteImage s = l.GetSprite() as ClippedSpriteImage;
            if (texture.SpriteImage == null)
            {
                texture.SpriteImage = new Image();
                texture.SpriteImage.Source = texture.Image;
                texture.SpriteImage.Opacity = 0;
                Canvas.Children.Add(texture.SpriteImage);
            }
            return s;
        }

        public void Draw(Texture2D texture, Vector2 position, Rectangle? sourceRectangle, Color color, float rotation, Vector2 origin, float scale, SpriteEffects effects, float layerDepth)
        {
            if (sourceRectangle == null)
            {
                sourceRectangle = texture.SourceRect;
            }

            ClippedSpriteImage d = GetClippedSprite(texture, color);
            if (d.ZIndex < zIndex)
            {
                d.ZIndex = zIndex;
            }
            else
            {
                zIndex = d.ZIndex;
            }
            zIndex++;
            d.InUse = true;
            d.Color = color;
            d.SourceRectangle = sourceRectangle.Value;
            d.Position = new Vector2(position.X, position.Y);
            d.Scale = new Vector2((float)scale, (float)scale);
            d.Rotation = rotation;
            d.Origin = origin;
            d.Effects = effects;
        }

        public void Draw(Texture2D texture, Vector2 position, Rectangle? sourceRectangle, Color color)
        {
            if (sourceRectangle == null)
            {
                sourceRectangle = texture.SourceRect;
            }

            ClippedSpriteImage d = GetClippedSprite(texture, color);
            if (d.ZIndex < zIndex)
            {
                d.ZIndex = zIndex;
            }
            else
            {
                zIndex = d.ZIndex;
            }
            zIndex++;
            d.InUse = true;
            d.Position = position;
            d.Color = color;
            d.SourceRectangle = sourceRectangle.Value;
            d.Scale = new Vector2(1, 1);
            d.Origin = Vector2.Zero;
            d.Effects = SpriteEffects.None;
        }

        public void Draw(Texture2D texture, Rectangle destinationRectangle, Rectangle? sourceRectangle, Color color)
        {
            if (sourceRectangle == null)
            {
                sourceRectangle = texture.SourceRect;
            }

            ClippedSpriteImage d = GetClippedSprite(texture, color);
            if (d.ZIndex < zIndex)
            {
                d.ZIndex = zIndex;
            }
            else
            {
                zIndex = d.ZIndex;
            }
            zIndex++;
            d.InUse = true;
            d.Color = color;
            d.SourceRectangle = sourceRectangle.Value;
            d.Position = new Vector2(destinationRectangle.X, destinationRectangle.Y);
            if (d.SourceRectangle.Width > 0 && d.SourceRectangle.Height > 0)
            {
                float scaleX = (float)destinationRectangle.Width / d.SourceRectangle.Width;
                float scaleY = (float)destinationRectangle.Height / d.SourceRectangle.Height;
                d.Scale = new Vector2((float)scaleX, (float)scaleY);
            }
            else
            {
                d.Scale = Vector2.Zero;
            }
            d.Origin = Vector2.Zero;
            d.Effects = SpriteEffects.None;
        }

        public void Draw(Texture2D texture, Rectangle destinationRectangle, Color color)
        {
            Draw(texture, destinationRectangle, texture.SourceRect, color);
        }

        public void Draw(Texture2D texture, Vector2 position, Color color)
        {
            Draw(texture, position, texture.SourceRect, color);
        }

        public void Draw(Texture2D texture, Rectangle destinationRectangle, Rectangle? sourceRectangle, Color color, float rotation, Vector2 origin, SpriteEffects effects, float layerDepth)
        {
            float sourceWidth = texture.Width;
            float sourceHeight = texture.Height;
            if (sourceRectangle != null)
            {
                sourceWidth = sourceRectangle.Value.Width;
                sourceHeight = sourceRectangle.Value.Height;
            }
            Vector2 scale = new Vector2(destinationRectangle.Width / sourceWidth, destinationRectangle.Height / sourceHeight);
            Vector2 pos = new Vector2(destinationRectangle.X, destinationRectangle.Y);
            Draw(texture, pos, sourceRectangle, color, rotation, origin, scale, effects, layerDepth);
        }

        public void Draw(Texture2D texture, Vector2 position, Rectangle? sourceRectangle, Color color, float rotation, Vector2 origin, Vector2 scale, SpriteEffects effects, float layerDepth)
        {
            if (sourceRectangle == null)
            {
                sourceRectangle = texture.SourceRect;
            }

            ClippedSpriteImage d = GetClippedSprite(texture, color);
            if (d.ZIndex < zIndex)
            {
                d.ZIndex = zIndex;
            }
            else
            {
                zIndex = d.ZIndex;
            }
            zIndex++;
            d.InUse = true;
            d.Color = color;
            d.SourceRectangle = sourceRectangle.Value;
            d.Position = new Vector2(position.X, position.Y);
            d.Scale = new Vector2((float)scale.X, (float)scale.Y);
            d.Rotation = rotation;
            d.Origin = origin;
            d.Effects = effects;
        }

    }
}
