﻿
/*
 * S3 Texture Compression (S3TC) decoding functions
 * Copyright (c) 2007 by Ivo van Poorten
 *
 * see also: http://wiki.multimedia.cx/index.php?title=S3TC
 *
 * The color decoding logic is ported by Bill Reiss to C# from the FFmpeg s3tc.c file.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Media.Imaging;
using System.IO;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
#if USE_FARSEER
using FarseerGames.FarseerPhysics.Mathematics;
#endif

namespace SilverArcade.SilverSprite.Graphics
{
    public class Texture2D : IDisposable
    {
        static int AssetCount;
        internal int AssetId;
        public string AssetName;
        int _rowLength;
        int _width = -1;
        int _height = -1;
        ImageSource imageSource;
        public Image SpriteImage;
        Color colorKey;
        internal byte[] _buffer;
        Rectangle rect = new Rectangle();
        static uint[] bitmasks = { 0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff };
        private bool _isDisposed = false;

        public event EventHandler Disposing;

        public Rectangle SourceRect
        {
            get
            {
                if (Width != rect.Width)
                {
                    rect.Width = Width;
                }
                if (Height != rect.Height)
                {
                    rect.Height = Height;
                }

                return rect;
            }
        }

        public Texture2D(Stream source, GraphicsDevice device)
        {
            _buffer = null;
            BitmapImage img = new BitmapImage();
            img.SetSource(source);
            imageSource = img;
            SpriteImage = new Image();
            SpriteImage.Source = img;
            SpriteImage.Opacity = 0;
            device.Root.Children.Add(SpriteImage);
            AssetId = AssetCount;
            AssetCount++;
            _width = -1;
            _height = -1;
        }

        public Texture2D(GraphicsDevice graphicsDevice, int width, int height, int numberLevels, TextureUsage usage, SurfaceFormat format)
        {
            _width = width;
            _height = height;
            rect.Width = _width;
            rect.Height = _height;
            _rowLength = _width * 4 + 1;
            _buffer = new byte[_rowLength * _height];
            AssetName = Guid.NewGuid().ToString();
            AssetId = AssetCount;
            AssetCount++;
        }

        public Texture2D(TextureCreationParameters param, byte[] data)
        {
            _width = param.Width;
            _height = param.Height;
            rect.Width = _width;
            rect.Height = _height;
            _rowLength = _width * 4 + 1;
            colorKey = param.ColorKey;
            SetData<byte>(data);
            AssetId = AssetCount;
            AssetCount++;
        }

        public Texture2D(Texture2D source, Color color)
        {
            if (source._buffer == null)
            {
                imageSource = source.imageSource;
                AssetName = source.AssetName;
                AssetId = AssetCount;
                AssetCount++;
                _width = source.Width;
                _height = source.Height;
                rect.Width = _width;
                rect.Height = _height;
                return;
            }
            _buffer = new byte[source._buffer.Length];
            _width = source.Width;
            _height = source.Height;
            _rowLength = _width * 4 + 1;
            rect.Width = _width;
            rect.Height = _height;
            source._buffer.CopyTo(_buffer, 0);
            AssetName = source.AssetName;
            for (int i = 0; i < _width; i++)
            {
                for (int j = 0; j < _height; j++)
                {
                    Color c = GetPixel(i, j);
                    SetPixel(i, j,
                        (byte)(c.R * (color.R / 255f)),
                        (byte)(c.G * (color.G / 255f)),
                        (byte)(c.B * (color.B / 255f)),
                        c.A);
                }
            }
            SetSource();
            AssetId = AssetCount;
            AssetCount++;
        }

        public Color GetPixel(int col, int row)
        {
            if (_buffer == null) return Color.Black;
            if ((col > _width) || (col < 0))
            {
        //        OnImageError("Error: Column must be greater than 0 and less than the Width");
            }
            else if ((row > _height) || (row < 0))
            {
        //        OnImageError("Error: Row must be greater than 0 and less than the Height");
            }

            Color color = new Color();
            int start = _rowLength * row + col * 4 + 1;   // +1 is for the filter byte

            color.R = _buffer[start];
            color.G = _buffer[start + 1];
            color.B = _buffer[start + 2];
            color.A = _buffer[start + 3];

            return color;
        }

        public void SetPixel(int col, int row, byte red, byte green, byte blue, byte alpha)
        {
            if (_buffer == null) return;

            if ((col > _width) || (col < 0))
            {
//                OnImageError("Error: Column must be greater than 0 and less than the Width");
            }
            else if ((row > _height) || (row < 0))
            {
//                OnImageError("Error: Row must be greater than 0 and less than the Height");
            }

            // Set the pixel
            int start = _rowLength * row + col * 4 + 1; // +1 is for the filter byte
            _buffer[start] = red;
            _buffer[start + 1] = green;
            _buffer[start + 2] = blue;
            _buffer[start + 3] = alpha;
        }

        void SetSource()
        {
            BitmapImage bmp = new BitmapImage();
            bmp.SetSource(Silverlight.Samples.PngEncoder.Encode(_buffer, _width, _height));
            imageSource = bmp;
        }

        public void SetData<T>(T[] data)
        {
            if (typeof(T) == typeof(byte))
            {
                byte[] b = data.Cast<byte>().ToArray();
                int rowLength = _width * 4 + 1;
                MemoryStream ms = new MemoryStream();
                BinaryWriter w = new BinaryWriter(ms);
                int idx = 0;
                for (int i = 0; i < _height; i++)
                {
                    w.Write((byte)0);
                    for (int j = 0; j < _width; j++)
                    {
                        w.Write(b[idx + 2]);
                        w.Write(b[idx + 1]);
                        w.Write(b[idx]);
                        w.Write(b[idx + 3]);
                        idx += 4;
                    }
                }
                _buffer = ms.GetBuffer();
                w.Close();
                ms.Close();
                ms.Dispose();
            }
            else if (typeof(T) == typeof(Color))
            {
                Color [] c = data.Cast<Color>().ToArray();
                int rowLength = _width * 4 + 1;
                MemoryStream ms = new MemoryStream();
                BinaryWriter w = new BinaryWriter(ms);
                int idx = 0;
                for (int i = 0; i < _height; i++)
                {
                    w.Write((byte)0);
                    for (int j = 0; j < _width; j++)
                    {
                        w.Write(c[idx].R);
                        w.Write(c[idx].G);
                        w.Write(c[idx].B);
                        w.Write(c[idx].A);
                        idx++;
                    }
                }
                _buffer = ms.GetBuffer();
                w.Close();
                ms.Close();
                ms.Dispose();
            }
            SetSource();

        }

        public ImageSource Image
        {
            get
            {
                return imageSource;
            }
            set
            {
                imageSource = value;
            }
        }

        public int Width
        {
            get
            {
                if (_width < 0)
                {
                    if (SpriteImage != null && SpriteImage.ActualWidth > 0)
                    {
                        _width = (int)SpriteImage.ActualWidth;
                        if (_height > 0)
                        {
                            SpriteImage.Visibility = Visibility.Collapsed;
                        }
                    }
                    else
                    {
                        return 0;
                    }
                }
                return _width;
            }
        }

        public int Height
        {
            get
            {
                if (_height < 0)
                {
                    if (SpriteImage != null && SpriteImage.ActualHeight > 0)
                    {
                        _height = (int)SpriteImage.ActualHeight;
                        if (_width > 0)
                        {
                            SpriteImage.Visibility = Visibility.Collapsed;
                        }
                    }
                    else
                    {
                        return 0;
                    }
                }
                return _height;
            }
        }

        static byte GetBits64(ulong source, int first, int length, int shift)
        {
            uint bitmask = bitmasks[length];
            source = source >> first;
            source = source & bitmask;
            source = source << shift;
            return (byte)source;
        }

        static byte GetBits(uint source, int first, int length, int shift)
        {
            uint bitmask = bitmasks[length];
            source = source >> first;
            source = source & bitmask;
            source = source << shift;
            return (byte)source;
        }

        static void SetColorFromPacked(byte[] data, int offset, byte alpha, uint packed)
        {
            byte r = (byte)(GetBits(packed, 0, 8, 0));
            byte g = (byte)(GetBits(packed, 8, 8, 0));
            byte b = (byte)(GetBits(packed, 16, 8, 0));
            data[offset] = r;
            data[offset + 1] = g;
            data[offset + 2] = b;
            data[offset + 3] = alpha;
        }

        static void ColorsFromPacked(uint[] colors, uint c0, uint c1, bool flag)
        {
            uint rb0, rb1, rb2, rb3, g0, g1, g2, g3;

            rb0  = (c0<<3 | c0<<8) & 0xf800f8;
            rb1  = (c1<<3 | c1<<8) & 0xf800f8;
            rb0 +=        (rb0>>5) & 0x070007;
            rb1 +=        (rb1>>5) & 0x070007;
            g0   =        (c0 <<5) & 0x00fc00;
            g1   =        (c1 <<5) & 0x00fc00;
            g0  +=        (g0 >>6) & 0x000300;
            g1  +=        (g1 >>6) & 0x000300;

            colors[0] = rb0 + g0;
            colors[1] = rb1 + g1;

            if (c0 > c1 || flag) {
                rb2 = (((2*rb0+rb1) * 21) >> 6) & 0xff00ff;
                rb3 = (((2*rb1+rb0) * 21) >> 6) & 0xff00ff;
                g2  = (((2*g0 +g1 ) * 21) >> 6) & 0x00ff00;
                g3  = (((2*g1 +g0 ) * 21) >> 6) & 0x00ff00;
                colors[3] = rb3 + g3;
            } else {
                rb2 = ((rb0+rb1) >> 1) & 0xff00ff;
                g2  = ((g0 +g1 ) >> 1) & 0x00ff00;
                colors[3] = 0;
            }

            colors[2] = rb2 + g2;
        }

        static Texture2D FromDxt1File(GraphicsDevice graphicsDevice, Stream stream, int length, TextureCreationParameters param)
        {
            BinaryReader rdr = new BinaryReader(stream);
            int xoffset = 0;
            int yoffset = 0;
            int rowLength = param.Width * 4;
            byte[] b = new byte[length];
            ushort c0;
            ushort c1;
            uint[] colors = new uint[4];

            uint lu;

            for (int y = 0; y < param.Height / 4; y++)
            {
                yoffset = y * 4;
                for (int x = 0; x < param.Width / 4; x++)
                {
                    xoffset = x * 4;
                    c0 = rdr.ReadUInt16();
                    c1 = rdr.ReadUInt16();
                    ColorsFromPacked(colors, c0, c1, false);

                    lu = rdr.ReadUInt32();
                    for (int i = 0; i < 16; i++)
                    {
                        int idx = GetBits(lu, 30 - i * 2, 2, 0);
                        int ii = 15 - i;
                        byte aa = 255;
                        int yy = yoffset + (ii / 4);
                        int xx = xoffset + (ii % 4);
                        int offset = yy * rowLength + xx * 4;
                        uint ci = colors[idx];
                        SetColorFromPacked(b, offset, aa, ci);
                    }
                }
            }
            return new Texture2D(param, b);
        }      
  
        static Texture2D FromDxt5File(GraphicsDevice graphicsDevice, Stream stream, int length, TextureCreationParameters param)
        {
            BinaryReader rdr = new BinaryReader(stream);
            int xoffset = 0;
            int yoffset = 0;
            int rowLength = param.Width * 4;
            byte[] b = new byte[length];
            ulong alpha;
            ushort c0;
            ushort c1;
            uint[] colors = new uint[4];

            uint lu;
            byte [] a = new byte[8];
            byte [] al = new byte[6];
            ulong ax;

            for (int y = 0; y < param.Height / 4; y++)
            {
                yoffset = y * 4;
                for (int x = 0; x < param.Width / 4; x++)
                {
                    xoffset = x * 4;
                    a[0] = rdr.ReadByte();
                    a[1] = rdr.ReadByte();
                    if (a[0] > a[1])
                    {
                        a[2] = (byte)(((6 * a[0]) + a[1]) / 7);
                        a[3] = (byte)(((5 * a[0]) + (2 * a[1])) / 7);
                        a[4] = (byte)(((4 * a[0]) + (3 * a[1])) / 7);
                        a[5] = (byte)(((3 * a[0]) + (4 * a[1])) / 7);
                        a[6] = (byte)(((2 * a[0]) + (5 * a[1])) / 7);
                        a[7] = (byte)(((1 * a[0]) + (6 * a[1])) / 7);
                    }
                    else
                    {
                        a[2] = (byte)(((4 * a[0]) + (1 * a[1])) / 5);
                        a[3] = (byte)(((3 * a[0]) + (2 * a[1])) / 5);
                        a[4] = (byte)(((2 * a[0]) + (3 * a[1])) / 5);
                        a[5] = (byte)(((1 * a[0]) + (2 * a[1])) / 5);
                        a[6] = 0;
                        a[7] = 255;
                    }
                    rdr.Read(al, 0, 6);
                    ax = (ulong)(
                        (((ulong)al[5]) << 40) + 
                        (((ulong)(al[4]) << 32) + 
                        (((ulong)al[3]) << 24) + 
                        (((ulong)al[2]) << 16) + 
                        (((ulong)al[1]) << 8) +
                        (ulong)al[0]));

                    c0 = rdr.ReadUInt16();
                    c1 = rdr.ReadUInt16();
                    ColorsFromPacked(colors, c0, c1, true);

                    lu = rdr.ReadUInt32();
                    for (int i = 0; i < 16; i++)
                    {
                        int idx = GetBits(lu, 30 - i * 2, 2, 0);
                        int ii = 15 - i;
                        int aidx = GetBits64(ax, ii * 3, 3, 0);
                        byte aa = a[aidx];
                        int yy = yoffset + (ii / 4);
                        int xx = xoffset + (ii % 4);
                        int offset = yy * rowLength + xx * 4;
                        uint ci = colors[idx];
                        SetColorFromPacked(b, offset, aa, ci);
                    }
                }
            }
            return new Texture2D(param, b);
        }      
  
        static Texture2D FromDxt3File(GraphicsDevice graphicsDevice, Stream stream, int length, TextureCreationParameters param)
        {
            BinaryReader rdr = new BinaryReader(stream);
            int xoffset = 0;
            int yoffset = 0;
            int rowLength = param.Width * 4;
            byte[] b = new byte[length];
            ulong alpha;
            ushort c0, c1;
            uint[] colors = new uint[4];
            uint lu;

            for (int y = 0; y < param.Height / 4; y++)
            {
                yoffset = y * 4;
                for (int x = 0; x < param.Width / 4; x++)
                {
                    xoffset = x * 4;
                    alpha = rdr.ReadUInt64();
                    c0 = rdr.ReadUInt16();
                    c1 = rdr.ReadUInt16();
                    ColorsFromPacked(colors, c0, c1, true);

                    lu = rdr.ReadUInt32();
                    for (int i = 0; i < 16; i++)
                    {
                        int idx = GetBits(lu, 30 - i * 2, 2, 0);
                        uint ci = colors[idx];
                        int ii = 15 - i;
                        byte a = (byte)(GetBits64(alpha, ii * 4, 4, 0));
                        a += (byte)(a << 4);
                        int yy = yoffset + (ii / 4);
                        int xx = xoffset + (ii % 4);
                        int offset = yy * rowLength + xx * 4;
                        SetColorFromPacked(b, offset, a, ci);
                    }
                }
            }
            return new Texture2D(param, b);
        }

        public static Texture2D FromFile(GraphicsDevice graphicsDevice, Stream stream, int length, TextureCreationParameters param)
        {
            if (param.Format == SurfaceFormat.Dxt3)
            {
                return FromDxt3File(graphicsDevice, stream, length, param);
            }
            else if (param.Format == SurfaceFormat.Dxt5)
            {
                return FromDxt5File(graphicsDevice, stream, length, param);
            }
            else if (param.Format == SurfaceFormat.Dxt1)
            {
                return FromDxt1File(graphicsDevice, stream, length, param);
            }
            else if (param.Format == SurfaceFormat.Color)
            {
                byte[] b = new byte[length];
                stream.Read(b, 0, length);
                Texture2D texture = new Texture2D(param, b);
                return texture;
            }
            else
            {
                return null;
            }
        }

        public SurfaceFormat Format
        {
            get { throw new NotImplementedException(); }
        }

        public TextureUsage TextureUsage
        {
            get { throw new NotImplementedException(); }
        }


        public Object Tag
        {
            get { throw new NotImplementedException(); }
            set { throw new NotImplementedException(); }
        }

        public bool IsDisposed
        {
            get { return _isDisposed; }
        }

        public string Name
        {
            get { throw new NotImplementedException(); }
        }

        public void Dispose()
        {
            //TODO: Any adidtional Cleanup Here?
            if(Disposing != null)
                Disposing(this, null);

            _isDisposed = true;
        }

        public static Texture2D FromFile(GraphicsDevice graphicsDevice, Stream textureStream)
        {
            throw new NotImplementedException();
        }

        public static Texture2D FromFile(GraphicsDevice graphicsDevice, Stream textureStream, int numberBytes)
        {
            throw new NotImplementedException();
        }

        public static Texture2D FromFile(GraphicsDevice graphicsDevice, Stream textureStream, TextureCreationParameters creationParameters)
        {
            throw new NotImplementedException();
        }

        public static Texture2D FromFile(GraphicsDevice graphicsDevice, string filename, TextureCreationParameters creationParameters)
        {
            throw new NotImplementedException();
        }

        public static Texture2D FromFile(GraphicsDevice graphicsDevice, string filename, int width, int height)
        {
            throw new NotImplementedException();
        }

        public static Texture2D FromFile(GraphicsDevice graphicsDevice, string filename)
        {
            throw new NotImplementedException();
        }

        public void SetData<T>(T[] data, int startIndex, int elementCount, SetDataOptions options)
		{
			throw new NotImplementedException();
		}

        public void SetData<T>(int level, Rectangle? rect, T[] data, int startIndex, int elementCount, SetDataOptions options)
		{
			throw new NotImplementedException();
		}

        public void GetData<T>(T[] data)
		{
			throw new NotImplementedException();
		}

		public void  GetData<T>(T[] data, int startIndex, int elementCount)
		{
			throw new NotImplementedException();
		}

        public void GetData<T>(int level, Rectangle? rect, T[] data, int startIndex, int elementCount)
		{
			throw new NotImplementedException();
		}

    }


}
